from __future__ import annotations

from pathlib import Path
from typing import Callable, Mapping, Sequence

from attrs import define

from jinjanator.cli import render_command


@define(kw_only=True)
class FilePair:
    template_file: Path
    data_file: Path


FilePairFactory = Callable[[str, str, str], FilePair]


def render_file(
    files: FilePair,
    options: Sequence[str],
) -> str:
    return render_command(
        Path.cwd(),
        {},
        None,
        ["", *options, str(files.template_file), str(files.data_file)],
    )


def render_implicit_stream(
    files: FilePair,
    options: Sequence[str],
) -> str:
    with files.data_file.open() as data:
        return render_command(
            Path.cwd(),
            {},
            data,
            ["", *options, str(files.template_file)],
        )


def render_explicit_stream(
    files: FilePair,
    options: Sequence[str],
) -> str:
    with files.data_file.open() as data:
        return render_command(
            Path.cwd(),
            {},
            data,
            ["", *options, str(files.template_file), "-"],
        )


def render_env(
    files: FilePair,
    options: Sequence[str],
    env: Mapping[str, str],
) -> str:
    return render_command(
        Path.cwd(),
        env,
        None,
        ["", *options, str(files.template_file)],
    )


def render_file_env(
    files: FilePair,
    options: Sequence[str],
    env: Mapping[str, str],
) -> str:
    return render_command(
        Path.cwd(),
        env,
        None,
        ["", *options, str(files.template_file), str(files.data_file)],
    )



from __future__ import annotations

from typing import TYPE_CHECKING

import pytest

from . import FilePair, FilePairFactory


if TYPE_CHECKING:
    from pathlib import Path


@pytest.fixture()
def make_file_pair(tmp_path: Path) -> FilePairFactory:
    def _make_file_pair(
        template_content: str,
        data_content: str,
        data_format: str,
    ) -> FilePair:
        template_file = tmp_path / "template.j2"
        template_file.write_text(template_content)
        data_file = tmp_path / f"data.{data_format}"
        data_file.write_text(data_content)
        return FilePair(template_file=template_file, data_file=data_file)

    return _make_file_pair



from __future__ import annotations

from . import (
    FilePairFactory,
    render_env,
    render_explicit_stream,
    render_file,
    render_implicit_stream,
)


DATA_TEMPLATE = """
server {
  listen 80;
  server_name {{ nginx.hostname }};

  root {{ nginx.webroot }};
  index index.htm;

  access_log {{ nginx.logs }}/http.access.log combined;
  error_log  {{ nginx.logs }}/http.error.log;
}
"""

ENV_TEMPLATE = """
server {
  listen 80;
  server_name {{ NGINX_HOSTNAME }};

  root {{ NGINX_WEBROOT }};
  index index.htm;

  access_log {{ NGINX_LOGS }}/http.access.log combined;
  error_log  {{ NGINX_LOGS }}/http.error.log;
}
"""

EXPECTED_OUTPUT = """
server {
  listen 80;
  server_name localhost;

  root /var/www/project;
  index index.htm;

  access_log /var/log/nginx/http.access.log combined;
  error_log  /var/log/nginx/http.error.log;
}
"""


def test_ini(make_file_pair: FilePairFactory) -> None:
    ini_data = """
[nginx]
hostname=localhost
webroot=/var/www/project
logs=/var/log/nginx
"""

    files = make_file_pair(DATA_TEMPLATE, ini_data, "ini")

    assert EXPECTED_OUTPUT == render_file(files, [])
    assert EXPECTED_OUTPUT == render_file(files, ["--format=ini"])

    assert EXPECTED_OUTPUT == render_implicit_stream(
        files,
        ["--format=ini"],
    )
    assert EXPECTED_OUTPUT == render_explicit_stream(
        files,
        ["--format=ini"],
    )


def test_json(make_file_pair: FilePairFactory) -> None:
    json_data = """
{
    "nginx":{
        "hostname": "localhost",
        "webroot": "/var/www/project",
        "logs": "/var/log/nginx"
    }
}
"""

    files = make_file_pair(DATA_TEMPLATE, json_data, "json")

    assert EXPECTED_OUTPUT == render_file(files, [])
    assert EXPECTED_OUTPUT == render_file(files, ["--format=json"])

    assert EXPECTED_OUTPUT == render_implicit_stream(
        files,
        ["--format=json"],
    )
    assert EXPECTED_OUTPUT == render_explicit_stream(
        files,
        ["--format=json"],
    )


def test_yaml(make_file_pair: FilePairFactory) -> None:
    yaml_data = """
nginx:
  hostname: localhost
  webroot: /var/www/project
  logs: /var/log/nginx
"""

    files = make_file_pair(DATA_TEMPLATE, yaml_data, "yaml")

    assert EXPECTED_OUTPUT == render_file(files, [])
    assert EXPECTED_OUTPUT == render_file(files, ["--format=yaml"])

    assert EXPECTED_OUTPUT == render_implicit_stream(
        files,
        ["--format=yaml"],
    )
    assert EXPECTED_OUTPUT == render_explicit_stream(
        files,
        ["--format=yaml"],
    )

    files = make_file_pair(DATA_TEMPLATE, yaml_data, "yml")

    assert EXPECTED_OUTPUT == render_file(files, [])


def test_env(make_file_pair: FilePairFactory) -> None:
    env_data = """
NGINX_HOSTNAME=localhost
NGINX_WEBROOT=/var/www/project
NGINX_LOGS=/var/log/nginx
"""

    files = make_file_pair(ENV_TEMPLATE, env_data, "env")

    assert EXPECTED_OUTPUT == render_file(files, [])
    assert EXPECTED_OUTPUT == render_file(files, ["--format=env"])

    assert EXPECTED_OUTPUT == render_explicit_stream(
        files,
        [],
    )
    assert EXPECTED_OUTPUT == render_explicit_stream(
        files,
        ["--format=env"],
    )

    env = {
        "NGINX_HOSTNAME": "localhost",
        "NGINX_WEBROOT": "/var/www/project",
        "NGINX_LOGS": "/var/log/nginx",
    }

    assert EXPECTED_OUTPUT == render_env(files, [], env)
    assert EXPECTED_OUTPUT == render_env(files, ["--format=env"], env)



from __future__ import annotations

import configparser
import json
import keyword

from typing import Any, Iterable, Mapping

import yaml

from jinjanator_plugins import (
    FormatOptionUnsupportedError,
    FormatOptionValueError,
    Formats,
    plugin_formats_hook,
)


class INIFormat:
    name = "ini"
    suffixes: Iterable[str] | None = (".ini",)
    option_names: Iterable[str] | None = ()

    def __init__(self, options: Iterable[str] | None) -> None:
        pass

    def parse(self, data_string: str) -> Mapping[str, Any]:
        """INI data input format.

        data.ini:

        ```
        [nginx]
        hostname=localhost
        webroot=/var/www/project
        logs=/var/log/nginx
        ```

        Usage:

        $ j2 config.j2 data.ini
        $ cat data.ini | j2 --format=ini config.j2
        """

        class MyConfigParser(configparser.ConfigParser):
            def as_dict(self) -> Mapping[str, Any]:
                d = dict(self._sections)  # type: ignore[attr-defined]
                for k in d:
                    d[k] = dict(self._defaults, **d[k])  # type: ignore[attr-defined]
                    d[k].pop("__name__", None)
                return d

        ini = MyConfigParser()
        ini.read_string(data_string)

        return ini.as_dict()


class JSONFormat:
    name = "json"
    suffixes: Iterable[str] | None = (".json",)
    option_names: Iterable[str] | None = "array-name"

    def __init__(self, options: Iterable[str] | None) -> None:
        self.array_name: str | None = None
        if options:
            for option in options:
                try:
                    opt, val = option.split("=")
                except ValueError as exc:
                    raise FormatOptionValueError(
                        self, option, "", "contains more than one '='"
                    ) from exc

                if not val.isidentifier():
                    raise FormatOptionValueError(
                        self, opt, val, "is not a valid Python identifier"
                    )

                if keyword.iskeyword(val):
                    raise FormatOptionValueError(self, opt, val, "is a Python keyword")

                self.array_name = val

    def parse(self, data_string: str) -> Mapping[str, Any]:
        """JSON data input format.

        data.json:

        ```
        {
            "nginx":{
                "hostname": "localhost",
                "webroot": "/var/www/project",
                "logs": "/var/log/nginx"
            }
        }
        ```

        Usage:

        $ j2 config.j2 data.json
        $ cat data.json | j2 --format=ini config.j2
        """

        try:
            context = json.loads(data_string)
        except json.decoder.JSONDecodeError as exc:
            msg = "JSON input is neither an object nor an array"
            raise TypeError(msg) from exc

        if isinstance(context, dict):
            if self.array_name:
                raise FormatOptionUnsupportedError(
                    self,
                    "array-name",
                    "cannot be used with object (dictionary) input",
                )

            return context

        if not self.array_name:
            raise FormatOptionUnsupportedError(
                self, "array-name", "must be specified for array (list) input"
            )

        return {self.array_name: context}


class YAMLFormat:
    name = "yaml"
    suffixes: Iterable[str] | None = (".yaml", ".yml")
    option_names: Iterable[str] | None = "sequence-name"

    def __init__(self, options: Iterable[str] | None) -> None:
        self.sequence_name: str | None = None
        if options:
            for option in options:
                try:
                    opt, val = option.split("=")
                except ValueError as exc:
                    raise FormatOptionValueError(
                        self, option, "", "contains more than one '='"
                    ) from exc

                if not val.isidentifier():
                    raise FormatOptionValueError(
                        self, opt, val, "is not a valid Python identifier"
                    )

                if keyword.iskeyword(val):
                    raise FormatOptionValueError(self, opt, val, "is a Python keyword")

                self.sequence_name = val

    def parse(
        self,
        data_string: str,
    ) -> Mapping[str, Any]:
        """YAML data input format.

        data.yaml:

        ```
        nginx:
            hostname: localhost
            webroot: /var/www/project
            logs: /var/log/nginx
        ```

        Usage:

        $ j2 config.j2 data.yml
        $ cat data.yml | j2 --format=yaml config.j2
        """

        context = yaml.safe_load(data_string)

        if isinstance(context, dict):
            if self.sequence_name:
                raise FormatOptionUnsupportedError(
                    self,
                    "sequence-name",
                    "cannot be used with mapping (dictionary) input",
                )

            return context

        if isinstance(context, list):
            if not self.sequence_name:
                raise FormatOptionUnsupportedError(
                    self, "sequence-name", "must be specified for sequence (array) input"
                )

            return {self.sequence_name: context}

        msg = "YAML input is neither a mapping nor a sequence"
        raise TypeError(msg)


class EnvFormat:
    name = "env"
    suffixes: Iterable[str] | None = (".env",)
    option_names: Iterable[str] | None = ()

    def __init__(self, options: Iterable[str] | None) -> None:
        pass

    def parse(self, data_string: str) -> Mapping[str, str]:
        """Data input from environment variables.

        Render directly from the current environment variable values:

        $ j2 config.j2

        Or alternatively, read the values from a dotenv file:

        ```
        NGINX_HOSTNAME=localhost
        NGINX_WEBROOT=/var/www/project
        NGINX_LOGS=/var/log/nginx/
        ```

        And render with:

        $ j2 config.j2 data.env
        $ env | j2 --format=env config.j2

        If you're going to pipe a dotenv file into `j2`, you'll need to
        use "-" as the second argument to explicitly:

        $ j2 config.j2 - < data.env
        """

        return dict(
            filter(
                lambda line: len(line) == 2,  # noqa: PLR2004
                (
                    list(map(str.strip, line.split("=", 1)))
                    for line in data_string.split("\n")
                ),
            ),
        )


@plugin_formats_hook
def plugin_formats() -> Formats:
    return {
        INIFormat.name: INIFormat,
        JSONFormat.name: JSONFormat,
        YAMLFormat.name: YAMLFormat,
        EnvFormat.name: EnvFormat,
    }



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Mapping, TextIO


if TYPE_CHECKING:  # pragma: no cover
    from jinjanator_plugins import (
        Format,
    )


def read_context_data(
    fmt: Format,
    f: TextIO | None,
    environ: Mapping[str, str],
    import_env: str | None = None,
) -> Mapping[str, Any]:
    if not f:
        msg = "no input supplied"
        raise ValueError(msg)

    context: dict[str, Any] = {}

    result = fmt.parse(f.read())

    context.update(result)

    if import_env is not None:
        if import_env == "":
            context.update(environ)
        else:
            context[import_env] = environ

    return context



from __future__ import annotations

import argparse
import importlib
import os
import sys

from pathlib import Path
from typing import (
    Any,
    Callable,
    Iterable,
    Mapping,
    Sequence,
    TextIO,
    cast,
)

import jinja2
import jinjanator_plugins
import pluggy

from . import filters, formats, version
from .context import read_context_data


class FilePathLoader(jinja2.BaseLoader):
    def __init__(self, cwd: Path, encoding: str = "utf-8"):
        self.cwd = cwd
        self.encoding = encoding

    def get_source(
        self,
        environment: jinja2.Environment,  # noqa: ARG002
        template_name: str,
    ) -> tuple[str, str, Callable[[], bool]]:
        template_path = Path(template_name)

        if not template_path.is_absolute():
            template_path = self.cwd / template_name

        if not template_path.is_file():
            raise jinja2.TemplateNotFound(template_name)

        mtime = template_path.stat().st_mtime

        return (
            template_path.read_text(encoding=self.encoding),
            str(template_path),
            lambda: template_path.stat().st_mtime == mtime,
        )


class Jinja2TemplateRenderer:
    ENABLED_EXTENSIONS = (
        "jinja2.ext.i18n",
        "jinja2.ext.do",
        "jinja2.ext.loopcontrols",
    )

    def __init__(
        self,
        cwd: Path,
        allow_undefined: bool,  # noqa: FBT001
        j2_env_params: dict[str, Any],
        plugin_hook_callers: jinjanator_plugins.PluginHookCallers,
    ):
        j2_env_params.setdefault("keep_trailing_newline", True)
        j2_env_params.setdefault(
            "undefined",
            jinja2.Undefined if allow_undefined else jinja2.StrictUndefined,
        )
        j2_env_params.setdefault("extensions", self.ENABLED_EXTENSIONS)
        j2_env_params.setdefault("loader", FilePathLoader(cwd))

        self._env = jinja2.Environment(**j2_env_params, autoescape=False)  # noqa: S701

        for plugin_globals in plugin_hook_callers.plugin_globals():
            self._env.globals.update(plugin_globals)

        for plugin_filters in plugin_hook_callers.plugin_filters():
            self._env.filters.update(plugin_filters)

        for plugin_tests in plugin_hook_callers.plugin_tests():
            self._env.tests.update(plugin_tests)

    def render(self, template_name: str, context: Mapping[str, str]) -> str:
        return self._env.get_template(template_name).render(context)


class UniqueStore(argparse.Action):
    """argparse action to restrict options to appearing only once."""

    def __init__(self, option_strings: Sequence[str], dest: str, **kwargs: Any) -> None:
        self.already_seen = False
        super().__init__(option_strings, dest, **kwargs)

    def __call__(
        self,
        parser: argparse.ArgumentParser,
        namespace: argparse.Namespace,
        values: str | Sequence[Any] | None,
        option_string: str | None = None,
    ) -> None:
        if self.already_seen and option_string:
            parser.error(option_string + " cannot be specified more than once.")
        setattr(namespace, self.dest, values)
        self.already_seen = True


def print_version_info(
    stream: TextIO = sys.stderr, *, plugin_identities: Iterable[str]
) -> None:
    print(
        f"{Path(sys.argv[0]).name} {version}, Jinja2"
        f" {importlib.metadata.version('jinja2')}",
        file=stream,
    )
    header_printed = False
    for plugin in plugin_identities:
        if not header_printed:
            print("Plugins:", file=stream)
            header_printed = True

        print(f"   {plugin}", file=stream)


class VersionAction(argparse.Action):
    def __init__(  # noqa: PLR0913
        self,
        option_strings: list[str],
        plugin_identities: Iterable[str],
        dest: str = argparse.SUPPRESS,
        default: str = argparse.SUPPRESS,
        help: str = "",  # noqa: A002
    ):
        super().__init__(
            option_strings=option_strings, dest=dest, default=default, nargs=0, help=help
        )
        self.plugin_identities = plugin_identities

    def __call__(
        self,
        parser: argparse.ArgumentParser,
        namespace: argparse.Namespace,  # noqa: ARG002
        values: str | Sequence[Any] | None,  # noqa: ARG002
        option_string: str | None = None,  # noqa: ARG002
    ) -> None:
        print_version_info(sys.stdout, plugin_identities=self.plugin_identities)
        parser.exit()


def parse_args(
    formats: Mapping[str, type[jinjanator_plugins.Format]],
    plugin_identities: Iterable[str],
    argv: Sequence[str] | None = None,
) -> argparse.Namespace:
    parser = argparse.ArgumentParser(
        prog="jinjanate",
        description="Command-line interface to Jinja2 for templating in shell scripts.",
        epilog="",
    )

    parser.add_argument(
        "-v",
        "--version",
        action=VersionAction,
        help="display version of this program and any installed plugins",
        plugin_identities=plugin_identities,
    )

    parser.add_argument(
        "-f",
        "--format",
        action=UniqueStore,
        default="?",
        help="Input data format",
        choices=["?", *list(formats.keys())],
    )

    parser.add_argument(
        "--format-option",
        action="append",
        metavar="option",
        dest="format_options",
        help="Options for data parser",
    )

    parser.add_argument(
        "-e",
        "--import-env",
        action=UniqueStore,
        default=None,
        metavar="VAR",
        dest="import_env",
        help=(
            "Import environment variables as `VAR` variable. Use empty string to import"
            " into the global scope"
        ),
    )

    parser.add_argument(
        "--undefined",
        action="store_true",
        dest="undefined",
        help="Allow undefined variables to be used in templates (suppress errors)",
    )

    parser.add_argument(
        "--quiet",
        action="store_true",
        dest="quiet",
        help="Suppress informational messages",
    )

    parser.add_argument(
        "-o",
        "--output-file",
        metavar="outfile",
        dest="output_file",
        type=Path,
        help="Output to a file instead of stdout",
    )

    parser.add_argument("template", help="Template file to process")

    parser.add_argument(
        "data",
        nargs="?",
        default=None,
        type=Path,
        help='Input data file name/path; "-" to use stdin',
    )

    return parser.parse_args(argv)


def get_hook_callers() -> jinjanator_plugins.PluginHookCallers:
    pm = pluggy.PluginManager("jinjanator")
    pm.add_hookspecs(jinjanator_plugins.PluginHooks)
    pm.register(filters)
    pm.register(formats)
    pm.load_setuptools_entrypoints("jinjanator")
    return cast(jinjanator_plugins.PluginHookCallers, pm.hook)


def validate_format_options(
    fmt: type[jinjanator_plugins.Format], options: Sequence[str] | None
) -> jinjanator_plugins.Format:
    if options:
        if not fmt.option_names:
            raise jinjanator_plugins.FormatOptionUnknownError(fmt, options[0])

        for opt in options:
            if opt.split("=")[0] not in fmt.option_names:
                raise jinjanator_plugins.FormatOptionUnknownError(fmt, opt)

    return fmt(options)


def render_command(
    cwd: Path,
    environ: Mapping[str, str],
    stdin: TextIO | None,
    argv: Sequence[str],
) -> str:
    plugin_hook_callers = get_hook_callers()

    available_formats: dict[str, type[jinjanator_plugins.Format]] = {}

    for plugin_formats in plugin_hook_callers.plugin_formats():
        available_formats.update(plugin_formats)

    plugin_identities = plugin_hook_callers.plugin_identities()

    args = parse_args(available_formats, plugin_identities, argv[1:])

    if not args.quiet:
        print_version_info(plugin_identities=plugin_identities)

    if args.format == "?":
        if args.data is None or str(args.data) == "-":
            args.format = "env"
        else:
            suffix = args.data.suffix
            for k, v in available_formats.items():
                if v.suffixes and suffix in v.suffixes:
                    args.format = k
                    break
            if args.format == "?":
                print(
                    f"No format which can read '{suffix}' files available",
                    file=sys.stderr,
                )
                raise SystemExit(1)

    # We always expect a file;
    # unless the user wants 'env', and there's no input file provided.
    if args.format == "env":
        """
        With the "env" format, if no dotenv filename is provided,
        we have two options: 1. The user wants to use the current
        environment 2. The user is feeding a dotenv file at stdin.
        Depending on whether we have data at stdin, we'll need to
        choose between the two.

        The problem is that in Linux, you can't reliably determine
        whether there is any data at stdin: some environments would
        open the descriptor even though they're not going to feed any
        data in.  That's why many applications would ask you to
        explicitly specify a '-' when stdin should be used.

        And this is what we're going to do here as well. The script,
        however, would give the user a hint that they should use '-'.
        """
        if str(args.data) == "-":
            input_data_f = stdin
        elif args.data is None:
            input_data_f = None
        else:
            input_data_f = args.data.open()
    else:
        input_data_f = (
            stdin if args.data is None or str(args.data) == "-" else args.data.open()
        )

    fmt = validate_format_options(available_formats[args.format], args.format_options)

    if args.format == "env" and input_data_f is None:
        context = environ
    else:
        context = read_context_data(
            fmt,
            input_data_f,
            environ,
            args.import_env,
        )

    renderer = Jinja2TemplateRenderer(
        cwd,
        args.undefined,
        j2_env_params={},
        plugin_hook_callers=plugin_hook_callers,
    )

    try:
        result = renderer.render(args.template, context)
    except jinja2.exceptions.UndefinedError as e:
        # When there's data at stdin, tell the user they should use '-'
        try:
            stdin_has_data = stdin is not None and not stdin.isatty()
            if args.format == "env" and args.data is None and stdin_has_data:
                extra_info = (
                    "\n\nIf you're trying to pipe a .env file, please run me with a '-'"
                    " as the data file name:\n$ {cmd} {argv} -".format(
                        cmd=Path(sys.argv[0]).name,
                        argv=" ".join(sys.argv[1:]),
                    )
                )
                e.args = (e.args[0] + extra_info,) + e.args[1:]
        except:  # noqa: E722, S110
            # The above code is so optional that any, ANY, error, is ignored
            pass

        # Proceed
        raise

    if args.output_file:
        with args.output_file.open("w") as f:
            f.write(result)
            f.close()
        return ""

    return result


def main(args: list[str] | None = None) -> int | None:
    try:
        if args is None:  # pragma: no cover
            args = sys.argv

        output = render_command(Path.cwd(), os.environ, sys.stdin, args)
    except jinjanator_plugins.FormatOptionUnknownError as exc:
        print(str(exc), file=sys.stderr)
        return 2
    except jinjanator_plugins.FormatOptionUnsupportedError as exc:
        print(str(exc), file=sys.stderr)
        return 3
    except jinjanator_plugins.FormatOptionValueError as exc:
        print(str(exc), file=sys.stderr)
        return 4
    except SystemExit as exc:
        if isinstance(exc.code, int):
            return exc.code

        return 1

    sys.stdout.write(output)

    return None



from __future__ import annotations

from pathlib import Path

import pytest

from jinja2.exceptions import TemplateNotFound

from . import (
    FilePairFactory,
    render_file,
)


def test_template_not_found(
    make_file_pair: FilePairFactory,
) -> None:
    files = make_file_pair("{{name}}", "", "env")
    files.template_file = Path("does-not-exist.j2")
    with pytest.raises(TemplateNotFound):
        render_file(files, [])



from __future__ import annotations

from typing import Iterable, Mapping

from jinjanator_plugins import (
    Filters,
    FormatOptionUnsupportedError,
    FormatOptionValueError,
    Formats,
    Globals,
    Identity,
    Tests,
    plugin_filters_hook,
    plugin_formats_hook,
    plugin_globals_hook,
    plugin_identity_hook,
    plugin_tests_hook,
)


def null_filter(
    value: str,  # noqa: ARG001
) -> str:
    return ""


def null_test(
    value: str,  # noqa: ARG001
) -> bool:
    return False


class NullFormat:
    name = "null"
    suffixes: Iterable[str] | None = (".null",)
    option_names: Iterable[str] | None = ("val", "uns")

    def __init__(self, options: Iterable[str] | None) -> None:
        if options:
            for option in options:
                if option == "val":
                    raise FormatOptionValueError(self, option, "", "")
                if option == "uns":
                    raise FormatOptionUnsupportedError(self, option, "")

    def parse(
        self,
        data_string: str,  # noqa: ARG002
    ) -> Mapping[str, str]:
        return {}


@plugin_identity_hook
def plugin_identities() -> Identity:
    return "test"


@plugin_filters_hook
def plugin_filters() -> Filters:
    return {"null": null_filter}


@plugin_tests_hook
def plugin_tests() -> Tests:
    return {"null": null_test}


@plugin_formats_hook
def plugin_formats() -> Formats:
    return {NullFormat.name: NullFormat}


@plugin_globals_hook
def plugin_globals() -> Globals:
    return {"null": null_filter}



from __future__ import annotations

import pytest

from jinja2.exceptions import UndefinedError

from . import (
    FilePairFactory,
    render_file,
)


def test_normal(make_file_pair: FilePairFactory) -> None:
    files = make_file_pair("{{name}}", "", "env")
    with pytest.raises(UndefinedError):
        render_file(files, [])


def test_suppressed(make_file_pair: FilePairFactory) -> None:
    files = make_file_pair("{{name}}", "", "env")
    assert "" == render_file(files, ["--undefined"])



from __future__ import annotations

import jinjanator_plugins
import pytest

from . import (
    FilePairFactory,
    render_file,
)


def test_ini_unknown(make_file_pair: FilePairFactory) -> None:
    files = make_file_pair("", "", "ini")

    with pytest.raises(jinjanator_plugins.FormatOptionUnknownError):
        assert render_file(files, ["--format-option", "midge"])


def test_json_unknown(make_file_pair: FilePairFactory) -> None:
    files = make_file_pair("", "", "json")

    with pytest.raises(jinjanator_plugins.FormatOptionUnknownError):
        assert render_file(files, ["--format-option", "midge"])


def test_yaml_unknown(make_file_pair: FilePairFactory) -> None:
    files = make_file_pair("", "", "yaml")

    with pytest.raises(jinjanator_plugins.FormatOptionUnknownError):
        assert render_file(files, ["--format-option", "midge"])


def test_env_unknown(make_file_pair: FilePairFactory) -> None:
    files = make_file_pair("", "", "env")

    with pytest.raises(jinjanator_plugins.FormatOptionUnknownError):
        assert render_file(files, ["--format-option", "midge"])



from __future__ import annotations

from typing import Any

import pytest

import jinjanator.cli

from . import (
    FilePairFactory,
    render_file,
)


def test_quiet(make_file_pair: FilePairFactory, capsys: Any) -> None:
    files = make_file_pair("Hello {{name}}!", "name=Blart", "env")
    render_file(files, ["--quiet"])
    captured = capsys.readouterr()
    assert 0 == len(captured.err)


def test_unavailable_suffix(make_file_pair: FilePairFactory, capsys: Any) -> None:
    files = make_file_pair("Hello {{name}}!", "name=Blart", "xyz")
    with pytest.raises(
        SystemExit,
    ):
        render_file(files, [])
    assert (
        "No format which can read '.xyz' files available"
        == capsys.readouterr().err.strip()
    )


def test_main_normal(make_file_pair: FilePairFactory, capsys: Any) -> None:
    files = make_file_pair("Hello {{name}}!", "name=Blart", "env")
    assert (
        jinjanator.cli.main(["", str(files.template_file), str(files.data_file)]) is None
    )
    assert "Hello Blart!" == capsys.readouterr().out


def test_main_failure(make_file_pair: FilePairFactory) -> None:
    files = make_file_pair("Hello {{name}}!", "name=Blart", "xyz")
    assert 1 == jinjanator.cli.main(["", str(files.template_file), str(files.data_file)])


def test_main_version(capsys: Any) -> None:
    """Ensure that the 'test' plugin's identity string is included in the '--version' output."""
    assert 0 == jinjanator.cli.main(["", "--version"])
    outlines = capsys.readouterr().out.splitlines()
    assert "Plugins:" == outlines[1]
    assert "   test" == outlines[2]


@pytest.mark.noplugin()
def test_main_version_no_plugins(capsys: Any) -> None:
    """Ensure that the 'test' plugin's identity string is not included in the '--version' output."""
    assert 0 == jinjanator.cli.main(["", "--version"])
    outlines = capsys.readouterr().out.splitlines()
    assert 1 == len(outlines)


def test_main_format_option_unknown(make_file_pair: FilePairFactory) -> None:
    files = make_file_pair("Hello {{name}}!", "name=Blart", "null")
    assert 2 == jinjanator.cli.main(  # noqa: PLR2004
        ["", "--format-option", "midge", str(files.template_file), str(files.data_file)]
    )


def test_main_format_option_unsupported(make_file_pair: FilePairFactory) -> None:
    files = make_file_pair("Hello {{name}}!", "name=Blart", "null")
    assert 3 == jinjanator.cli.main(  # noqa: PLR2004
        ["", "--format-option", "uns", str(files.template_file), str(files.data_file)]
    )


def test_main_format_option_value_error(make_file_pair: FilePairFactory) -> None:
    files = make_file_pair("Hello {{name}}!", "name=Blart", "null")
    assert 4 == jinjanator.cli.main(  # noqa: PLR2004
        ["", "--format-option", "val", str(files.template_file), str(files.data_file)]
    )



from __future__ import annotations

from typing import Iterable, Mapping

import pytest

from jinjanator.cli import parse_args


class FakeFormat:
    name = "env"
    suffixes: Iterable[str] | None = (".env",)
    option_names: Iterable[str] | None = ()

    def __init__(self, options: Iterable[str] | None) -> None:
        pass

    def parse(
        self,
        data_string: str,  # noqa: ARG002
    ) -> Mapping[str, str]:
        return {"foo": "bar"}


def test_invalid_arg() -> None:
    """Ensure that an invalid argument is not accepted."""
    with pytest.raises(SystemExit):
        parse_args({}, [], ["--test-invalid-arg"])


@pytest.mark.parametrize(
    "args",
    [
        ["--format", "env"],
        ["--format-option", "opt"],
        ["--import-env", "env"],
        ["--output-file", "output"],
        ["--quiet"],
        ["--undefined"],
        ["-e", "env"],
        ["-f", "env"],
        ["-o", "output"],
    ],
)
def test_args(args: list[str]) -> None:
    """Ensure that known arguments are accepted."""
    parse_args({"env": FakeFormat}, [], [*args, "template"])


def test_version() -> None:
    """Ensure that '--version' argument is accepted and program exits without an error."""
    with pytest.raises(SystemExit) as excinfo:
        parse_args({}, [], ["--version"])
    assert 0 == excinfo.value.code


@pytest.mark.xfail()
@pytest.mark.parametrize(
    "args",
    [
        ["--format", "env", "-f", "env"],
        ["--import-env", "env", "-e", "env"],
    ],
)
def test_duplicate_args(args: list[str]) -> None:
    """Ensure that duplicate arguments are not accepted."""
    parse_args({"env": FakeFormat}, [], [*args, "template"])



from __future__ import annotations

import os

from typing import Any

import pytest

from jinja2 import UndefinedError

from . import (
    FilePairFactory,
    render_file,
    render_file_env,
    render_implicit_stream,
)


def test_import(make_file_pair: FilePairFactory) -> None:
    # Import environment into a variable
    files = make_file_pair("{{ a }}/{{ env.B }}", '{"a":1}', "json")
    assert "1/2" == render_file_env(files, ["--import-env=env"], env={"B": "2"})

    # Import environment into global scope
    files = make_file_pair("{{ a }}/{{ B }}", '{"a":1, "B": 1}', "json")
    assert "1/2" == render_file_env(files, ["--import-env="], env={"B": "2"})


def test_equals_sign_in_file_value(make_file_pair: FilePairFactory) -> None:
    files = make_file_pair(
        "{{ A|default() }}/{{ B }}/{{ C }}",
        "A\nB=1\nC=val=1\n",
        "env",
    )
    assert "/1/val=1" == render_file(files, [])


def test_filter(make_file_pair: FilePairFactory, monkeypatch: Any) -> None:
    assert "USER_PASS" not in os.environ

    files = make_file_pair(
        '{{ user_login }}:{{ "USER_PASS"|env }}',
        "user_login: kolypto",
        "yaml",
    )

    # Value provided by environment
    monkeypatch.setenv("USER_PASS", "qwerty123")
    assert "kolypto:qwerty123" == render_file(files, [])

    # Value not provided
    monkeypatch.delenv("USER_PASS")
    with pytest.raises(KeyError):
        assert "kolypto:qwerty123" == render_file(files, [])

    # Default value
    files = make_file_pair(
        '{{ user_login }}:{{ "USER_PASS"|env("-none-") }}',
        "user_login: kolypto",
        "yaml",
    )

    assert "kolypto:-none-" == render_file(files, [])


def test_function(make_file_pair: FilePairFactory, monkeypatch: Any) -> None:
    assert "USER_PASS" not in os.environ

    files = make_file_pair(
        '{{ user_login }}:{{ env("USER_PASS") }}',
        "user_login: kolypto",
        "yaml",
    )

    # Value provided by environment
    monkeypatch.setenv("USER_PASS", "qwerty123")
    assert "kolypto:qwerty123" == render_file(files, [])

    # Value not provided
    monkeypatch.delenv("USER_PASS")
    with pytest.raises(KeyError):
        assert "kolypto:qwerty123" == render_file(files, [])

    # Default value
    files = make_file_pair(
        '{{ user_login }}:{{ env("USER_PASS", "-none-") }}',
        "user_login: kolypto",
        "yaml",
    )
    assert "kolypto:-none-" == render_file(files, [])


def test_env_stream(make_file_pair: FilePairFactory) -> None:
    files = make_file_pair("{{ a }}", "foo=bar", "env")
    with pytest.raises(UndefinedError, match="If you're trying to pipe a .env file"):
        render_implicit_stream(files, ["--format=env"])



from __future__ import annotations

import jinjanator_plugins
import pytest

from . import (
    FilePairFactory,
    render_file,
)


def test_mapping_normal(make_file_pair: FilePairFactory) -> None:
    files = make_file_pair("{{ foo }}", '{"foo": "bar"}', "json")

    assert "bar" == render_file(files, [])


def test_mapping_with_array_name_option(make_file_pair: FilePairFactory) -> None:
    files = make_file_pair("{{ foo }}", '{"foo": "bar"}', "json")

    with pytest.raises(jinjanator_plugins.FormatOptionUnsupportedError):
        assert render_file(files, ["--format-option", "array-name=seq"])


def test_array_normal(make_file_pair: FilePairFactory) -> None:
    files = make_file_pair("{{ seq[0] }}", "[1,2,3]", "json")

    assert "1" == render_file(files, ["--format-option", "array-name=seq"])


def test_array_without_array_name_option(make_file_pair: FilePairFactory) -> None:
    files = make_file_pair("{{ seq[0] }}", "[1,2,3]", "json")

    with pytest.raises(jinjanator_plugins.FormatOptionUnsupportedError):
        assert render_file(files, [])


def test_array_invalid_name(make_file_pair: FilePairFactory) -> None:
    files = make_file_pair("{{ seq[0] }}", "[1,2,3]", "json")

    with pytest.raises(jinjanator_plugins.FormatOptionValueError):
        render_file(files, ["--format-option", "array-name=334seq"])


def test_array_invalid_value(make_file_pair: FilePairFactory) -> None:
    files = make_file_pair("{{ seq[0] }}", "[1,2,3]", "json")

    with pytest.raises(jinjanator_plugins.FormatOptionValueError):
        render_file(files, ["--format-option", "array-name=abc=def"])


def test_array_keyword_name(make_file_pair: FilePairFactory) -> None:
    files = make_file_pair("{{ seq[0] }}", "[1,2,3]", "json")

    with pytest.raises(jinjanator_plugins.FormatOptionValueError):
        render_file(files, ["--format-option", "array-name=raise"])



from __future__ import annotations

from . import (
    FilePairFactory,
    render_file,
)


def test_extension(make_file_pair: FilePairFactory) -> None:
    files = make_file_pair("{% do [] %}", "", "env")
    assert "" == render_file(files, [])



from __future__ import annotations

from . import (
    FilePairFactory,
    render_env,
    render_file,
)


def test_json(make_file_pair: FilePairFactory) -> None:
    # I'm using Russian language for unicode :)
    files = make_file_pair(
        "Проверка {{ a }} связи!",
        '{"a": "широкополосной"}',
        "json",
    )
    assert "Проверка широкополосной связи!" == render_file(files, [])


def test_env(make_file_pair: FilePairFactory) -> None:
    files = make_file_pair("Hello {{name}}!", "", "env")
    # Test case from issue #17 (in j2cli repo): unicode environment variables
    assert "Hello Jürgen!" == render_env(files, [], env={"name": "Jürgen"})



from __future__ import annotations

import pytest

from . import (
    FilePairFactory,
    render_env,
    render_file,
)


def test_invalid_json(make_file_pair: FilePairFactory) -> None:
    files = make_file_pair("", "midge", "json")

    with pytest.raises(TypeError, match="JSON input is neither an object nor an array"):
        assert render_file(files, [])


def test_invalid_yaml(make_file_pair: FilePairFactory) -> None:
    files = make_file_pair("", "midge", "yaml")

    with pytest.raises(TypeError, match="YAML input is neither a mapping nor a sequence"):
        assert render_file(files, [])


def test_no_input_data(make_file_pair: FilePairFactory) -> None:
    files = make_file_pair("", "", "yaml")

    with pytest.raises(ValueError, match="no input supplied"):
        assert render_env(files, ["--format", "yaml"], {})



from __future__ import annotations

import os

from jinjanator_plugins import (
    Filters,
    Globals,
    plugin_filters_hook,
    plugin_globals_hook,
)


def env(varname: str, default: str | None = None) -> str:
    """Use an environment variable's value inside your template.

    This filter is available even when your data source is something other that the environment.

    Example:
    -------
    ```jinja2
    User: {{ user_login }}
    Pass: {{ "USER_PASSWORD"|env }}
    ```

    You can provide the default value:

    ```jinja2
    Pass: {{ "USER_PASSWORD"|env("-none-") }}
    ```

    For your convenience, it's also available as a function:

    ```jinja2
    User: {{ user_login }}
    Pass: {{ env("USER_PASSWORD") }}
    ```

    Notice that there must be quotes around the environment variable name
    """
    if default is not None:
        # With the default, there's never an error
        return os.getenv(varname, default)

    # Raise KeyError when not provided
    return os.environ[varname]


@plugin_filters_hook
def plugin_filters() -> Filters:
    return {"env": env}


@plugin_globals_hook
def plugin_globals() -> Globals:
    return {"env": env}



from __future__ import annotations

from typing import TYPE_CHECKING

from . import (
    FilePairFactory,
    render_env,
)


if TYPE_CHECKING:
    import pathlib


def test_short_option(
    make_file_pair: FilePairFactory,
    tmp_path: pathlib.Path,
) -> None:
    files = make_file_pair("{{ a }}", "", "json")
    out_file = tmp_path / "j2-out"
    assert "" == render_env(files, ["-o", str(out_file)], env={"a": "123"})
    assert "123" == out_file.read_text()


def test_long_option(
    make_file_pair: FilePairFactory,
    tmp_path: pathlib.Path,
) -> None:
    files = make_file_pair("{{ a }}", "", "json")
    out_file = tmp_path / "j2-out"
    assert "" == render_env(files, ["--output-file", str(out_file)], env={"a": "123"})
    assert "123" == out_file.read_text()
